// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ResidualEstimator.h */

/** \defgroup Estimator Estimator module
 * @{ <img src="estimator.png"> @}
 */

#ifndef AMDIS_RESIDUALESTIMATOR_H
#define AMDIS_RESIDUALESTIMATOR_H

#include "Estimator.h"
#include "FixVec.h"

namespace AMDiS {

  /** \brief
   * Returns residual square at quadrature point. Not Member of
   * Estimator to avoid multiple instantiation.
   */
  void r(const ElInfo *elInfo,
	 int nPoints,
	 const double *uhIq,
	 const WorldVector<double> *grdUhIq,
	 const WorldMatrix<double> *D2UhIq,
	 const double *uhOldIq,
	 const WorldVector<double> *grdUhOldIq,
	 const WorldMatrix<double> *D2UhOldIq,
	 DOFMatrix *A, 
	 DOFVector<double> *fh,
	 Quadrature *quad,
	 double *result);
 
  /** \brief
   * Returns pow(det,2.0/dim). Not Member of
   * Estimator to avoid multiple instantiation.
   */
  inline double h2_from_det(double det, int dim) {
    return pow(det, 2.0 / dim);
  }

  // ============================================================================
  // ===== class ResidualEstimator ==============================================
  // ============================================================================

  /**
   * \ingroup Estimator
   * 
   * \brief
   * Estimator for scalar problems.
   */
  class ResidualEstimator : public Estimator
  {
  public:
    /** \brief
     * Creator class used in the OEMSolverMap.
     */
    class Creator : public EstimatorCreator
    {
    public:
      Creator() : EstimatorCreator() {}

      virtual ~Creator() {}

      /** \brief
       * Returns a new ODirSolver object.
       */   
      Estimator* create() { 
	return new ResidualEstimator(name, row);
      }
    };
  

    /** \brief
     * Constructor.
     */
    ResidualEstimator(std::string name, int r);

    virtual void init(double timestep);

    virtual void estimateElement(ElInfo *elInfo);

    virtual void exit(bool output = true);

  protected:
    /** \brief
     * Constant in front of element residual
     */
    double C0;
 
    /** \brief
     * Constant in front of edge/face residual
     */
    double C1;

    /** \brief
     * Constant in front of coarsening term
     */
    double C2;

    /** \brief
     * Constant in front of the time
     */
    double C3;

    /// Number of systems, e.g., number of variables in the equation.
    int nSystems;

    /// Number of quadrature points.
    int nPoints;

    int dim;

    int degree;

    Quadrature *quad;

    FastQuadrature **quadFast;

    const BasisFunction **basFcts;

    double **uhEl;

    double **uhOldEl;

    double **uhNeigh;

    double *uhQP;

    double *uhOldQP;

    double *riq;

    WorldVector<double> *grdUh_qp;

    WorldMatrix<double> *D2uhqp; 

    ElInfo *neighInfo;

    Quadrature *surfaceQuad_;

    int nPointsSurface_;

    Vector<WorldVector<double> > grdUhEl_;

    Vector<WorldVector<double> > grdUhNeigh_;

    Vector<WorldVector<double> > jump_;

    Vector<WorldVector<double> > localJump_;

    WorldVector<int> faceIndEl_;
    
    WorldVector<int> faceIndNeigh_;

    DimVec<WorldVector<double> > *lambdaNeigh_;

    DimVec<double> *lambda_;

    int neighbours_;
  };
}

#endif // AMDIS_RESIDUALESTIMATOR_H
